"""A set of utilities designed to assist with handling and debugging NaNs in
pytorch.

* `isnan()` is an element-wise NaN check for pytorch Tensors.
* `hasnan()` determines whether or not a pytorch Tensor has any NaNs.
* `findnan()` will attempt to find NaNs in arbitrarily nested Python objects.

This module also offers a wrapped version of the entire `torch` module which
checks for NaNs in all inputs and outputs from all `torch` functions. For
example,

```
In [1]: from kindling.nan_police import torch

In [2]: x = torch.ones(2) / 0 * 0

In [3]: x
Out[3]:

nan
nan
[torch.FloatTensor of size 2]

In [4]: torch.sum(x)
---------------------------------------------------------------------------
Exception                                 Traceback (most recent call last)
<ipython-input-4-e7f45fec8fb4> in <module>()
----> 1 torch.sum(x)

~/Development/kindling/kindling/nan_police.py in __call__(self, *args, **kwargs)
    147         if argnan_path == []:
    148           raise Exception(
--> 149             f'Found a NaN at positional argument {i + 1} (of {len(args)}) when '
    150             f'calling `{path}`!'
    151           )

Exception: Found a NaN at positional argument 1 (of 1) when calling `torch.sum`!
```

Just import `from kindling.nan_police import torch` in all of your modules when
you'd like to test for NaNs! Note that this won't inject itself into all
references to torch, only those where you've imported this wrapped version of
pytorch. That means that it won't affect the behavior of torch in third-party
modules for example.
"""

import math
import numbers

import numpy as np
import torch as realtorch


def isnan(tensor):
  # Gross: https://github.com/pytorch/pytorch/issues/4767
  return (tensor != tensor)

def hasnan(tensor):
  return isnan(tensor).any()

def findnan(obj):
  """Find a NaN in the haystack of `obj` and return the path to finding it.
  Returns None if no NaNs could be found. Works with Python numbers, lists,
  tuples, Pytorch Tensors, Variables, and numpy arrays."""
  # Python numbers
  if isinstance(obj, numbers.Number):
    return ([] if math.isnan(obj) else None)

  # Pytorch Tensors
  elif isinstance(obj, realtorch.Tensor):
    return ([] if hasnan(obj) else None)

  # Pytorch Variables
  elif isinstance(obj, realtorch.autograd.Variable):
    if hasnan(obj.data):
      return [('Variable.data',)]
    else:
      gradnan_path = findnan(obj.grad)
      if gradnan_path == None:
        return None
      else:
        return [('Variable.grad',)] + gradnan_path

  # numpy array (See https://stackoverflow.com/questions/36783921/valueerror-when-checking-if-variable-is-none-or-numpy-array)
  elif isinstance(obj, np.ndarray):
    return ([] if np.isnan(obj).any() else None)

  # Python list
  elif isinstance(obj, list):
    for ix, elem in enumerate(obj):
      elemnan_path = findnan(elem)
      if elemnan_path != None:
        return [('list[]', ix)] + elemnan_path

    # Haven't found any NaNs yet, we're done
    return None

  # Python tuple
  elif isinstance(obj, tuple):
    for ix, elem in enumerate(obj):
      elemnan_path = findnan(elem)
      if elemnan_path != None:
        return [('tuple[]', ix)] + elemnan_path

    # Haven't found any NaNs yet, we're done
    return None

  # Python dict
  elif isinstance(obj, dict):
    for k, v in obj.items():
      elemnan_path = findnan(v)
      if elemnan_path != None:
        return [('dict[]', k)] + elemnan_path

    # Haven't found any NaNs yet, we're done
    return None

  # Don't know what we're looking at, assume it doesn't have NaNs.
  else:
    return None

def path_to_string(path):
  """Convert a path found with `findnan` to a pseudo-code selector."""
  if len(path) == 0:
    return ''
  else:
    first = path[0]
    rest = path[1:]
    typ = first[0]
    if typ == 'Variable.data':
      return '.data' + path_to_string(rest)
    elif typ == 'Variable.grad':
      return '.grad' + path_to_string(rest)
    elif (typ == 'list[]') or (typ == 'tuple[]') or (typ == 'dict[]'):
      return f'[{first[1]}]' + path_to_string(rest)
    else:
      raise Exception(f'Unrecognized path type, {typ}')

class Mock(object):
  """Emulate a module or function and wrap all calls with NaN checking. Throws
  an Exception when any NaNs are found."""
  def __init__(self, obj, path):
    self._obj = obj
    self._path = path

  def __call__(self, *args, **kwargs):
    path = '.'.join(self._path)

    # check args
    for i, arg in enumerate(args):
      argnan_path = findnan(arg)
      if argnan_path != None:
        if argnan_path == []:
          raise Exception(
            f'Found a NaN at positional argument {i + 1} (of {len(args)}) when '
            f'calling `{path}`!'
          )
        else:
          raise Exception(
            f'Found NaN in positional argument {i + 1} (of {len(args)}) when '
            f'calling `{path}`! Specifically at '
            f'`<arg>{path_to_string(argnan_path)}`.'
          )

    # check kwargs
    for k, v in kwargs.items():
      vnan_path = findnan(v)
      if vnan_path != None:
        if vnan_path == []:
          raise Exception(
            f'Found a NaN at keyword argument `{k}` when calling `{path}`!'
          )
        else:
          raise Exception(
            f'Found NaN in keyword argument `{k}` when calling `{path}`! '
            f'Specifically at `<{k}>{path_to_string(vnan_path)}`.'
          )

    result = self._obj(*args, **kwargs)

    # check result for NaNs
    resultnan_path = findnan(result)
    if resultnan_path != None:
      if resultnan_path == []:
        raise Exception(f'Found NaN in output from `{path}`!')
      else:
        raise Exception(
          f'Found NaN in output from `{path}`! Specifically at '
          f'`<out>{path_to_string(resultnan_path)}`.'
        )

    return result

  def __getattr__(self, name):
    return Mock(getattr(self._obj, name), self._path + [name])

  def __dir__(self):
    return self._obj.__dir__()

torch = Mock(realtorch, ['torch'])



import torch
from torch.autograd import Variable

from .utils import KL_Normals


class VAE(object):
  """An implementation of a variational autoenoder. Note that this particular
  implementation requires that the prior over z and the approximate posterior
  q(z | x) must both be diagonal normal distributions."""

  def __init__(
      self,
      inference_model,
      generative_model,
      prior_z
  ):
    self.inference_model = inference_model
    self.generative_model = generative_model
    self.prior_z = prior_z

  def evaluate_elbo(self, X, mc_samples):
    batch_size = X.size(0)

    # [batch_size, dim_z]
    q_z = self.inference_model(X)

    # KL divergence is additive across independent joint distributions, so this
    # works appropriately. This is where the normal constraints come from.
    z_kl = KL_Normals(q_z, self.prior_z.expand_as(q_z)) / batch_size

    # [batch_size * mc_samples, dim_z]
    z_sample = torch.cat([q_z.sample() for _ in range(mc_samples)], dim=0)
    Xrep = Variable(X.data.repeat(mc_samples, 1))
    reconstr = self.generative_model(z_sample)
    loglik_term = (
      reconstr.logprob(Xrep)
      / mc_samples
      / batch_size
    )

    elbo = -z_kl + loglik_term

    return {
      'elbo': elbo,
      'z_kl': z_kl,
      'reconstruction_log_likelihood': loglik_term,
      'z_sample': z_sample,
      'reconstruction': reconstr
    }



import itertools
import math

import torch
from torch.autograd import Variable
from torch.nn import functional as F

# from .utils import split


LOG2PI = torch.log(torch.FloatTensor([2 * math.pi]))[0]

# class NormalLogStdDev(object):
#   def __init__(self, mean, log_stddev):
#     assert mean.size() == log_stddev.size()
#     self.mean = mean
#     self.log_stddev = log_stddev

#   def sample(self):
#     return self.mean + torch.exp(self.log_stddev) * Variable(torch.randn(self.mean.size()))

#   def logprob(self, x):
#     return -0.5 * (
#       self.mean.numel() * LOG2PI
#       + 2 * torch.sum(self.log_stddev)
#       + torch.sum((x - self.mean) * torch.exp(-2 * self.log_stddev) * (x - self.mean))
#     )

class Normal(object):
  def __init__(self, mu, sigma):
    assert mu.size() == sigma.size()
    self.mu = mu
    self.sigma = sigma

  def size(self, *args, **kwargs):
    return self.mu.size(*args, **kwargs)

  def sample(self):
    eps = torch.randn(self.mu.size()).type_as(self.mu.data)
    return self.mu + self.sigma * Variable(eps)

  def logprob(self, x):
    return torch.sum(self.logprob_independent(x))

  def logprob_independent(self, x):
    return (
      -0.5 * LOG2PI
      - torch.log(self.sigma)
      -0.5 * torch.pow((x - self.mu) / self.sigma, 2)
    )

  def expand(self, *args, **kwargs):
    return Normal(
      self.mu.expand(*args, **kwargs),
      self.sigma.expand(*args, **kwargs)
    )

  def expand_as(self, *args, **kwargs):
    return Normal(
      self.mu.expand_as(*args, **kwargs),
      self.sigma.expand_as(*args, **kwargs)
    )

class Bernoulli(object):
  def __init__(self, rate):
    self.rate = rate

  def logprob(self, x):
    return -F.binary_cross_entropy(self.rate, x, size_average=False)

  def logprob_independent(self, x):
    return x * torch.log(self.rate) + (1 - x) * torch.log(1 - self.rate)

class Gamma(object):
  def __init__(self, shape, rate):
    assert shape.size() == rate.size()
    self.shape = shape
    self.rate = rate

  def logprob(self, x):
    return torch.sum(
      self.shape * torch.log(self.rate)
      - torch.lgamma(self.shape)
      + (self.shape - 1) * torch.log(x)
      - self.rate * x
    )

class NormalNet(object):
  def __init__(self, mu_net, sigma_net):
    self.mu_net = mu_net
    self.sigma_net = sigma_net

  def __call__(self, x):
    return Normal(self.mu_net(x), self.sigma_net(x))

  def parameters(self):
    return itertools.chain(
      self.mu_net.parameters(),
      self.sigma_net.parameters()
    )

  def cuda(self):
    self.mu_net.cuda()
    self.sigma_net.cuda()

class BernoulliNet(object):
  def __init__(self, rate_net):
    self.rate_net = rate_net

  def __call__(self, x):
    return Bernoulli(self.rate_net(x))

  def parameters(self):
    return self.rate_net.parameters()

# class _DistributionCat(object):
#   def __init__(self, distributions, dim=-1):
#     self.distributions = distributions
#     self.dim = dim

#   def sample(self):
#     return torch.cat([d.sample() for d in self.distributions], dim=self.dim)

#   def logprob(self, x):
#     xs = split(x, [d.size(self.dim) for d in self.distributions], dim=self.dim)
#     return sum(d.logprob(xx) for d, xx in zip(self.distributions, xs))

#   def entropy(self):
#     return sum(d.entropy() for d in self.distributions)

# def DistributionCat(distributions, dim=-1):
#   if all(isinstance(d, Normal) for d in distributions):
#     return Normal(
#       mu=torch.cat([d.mu for d in distributions], dim=dim),
#       sigma=torch.cat([d.sigma for d in distributions], dim=dim)
#     )
#   else:
#     return _DistributionCat(distributions, dim=-1)



import torch


def KL_Normals(d0, d1):
  return torch.sum(KL_Normals_independent(d0, d1))

# See https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence#Multivariate_normal_distributions
def KL_Normals_independent(d0, d1):
  assert d0.mu.size() == d1.mu.size()
  sigma0_sqr = torch.pow(d0.sigma, 2)
  sigma1_sqr = torch.pow(d1.sigma, 2)
  return (
    -0.5
    + (sigma0_sqr + torch.pow(d0.mu - d1.mu, 2)) / (2 * sigma1_sqr)
    + torch.log(d1.sigma)
    - torch.log(d0.sigma)
  )

class Lambda(torch.nn.Module):
  def __init__(self, func, extra_args=(), extra_kwargs={}):
    super(Lambda, self).__init__()
    self.func = func
    self.extra_args = extra_args
    self.extra_kwargs = extra_kwargs

  def forward(self, x):
    return self.func(x, *self.extra_args, **self.extra_kwargs)

class MetaOptimizer(object):
  def __init__(self, optimizers):
    self.optimizers = optimizers

  def zero_grad(self):
    for opt in self.optimizers:
      opt.zero_grad()

  def step(self):
    for opt in self.optimizers:
      opt.step()



from setuptools import setup

import kindling


setup(
  name='kindling',
  version=kindling.__version__,
  description='Basic building blocks for pytorch.',
  author='Samuel Ainsworth',
  author_email='skainsworth@gmail.com',
  url='https://github.com/samuela/kindling',
  packages=['kindling'],
  install_requires=[]
)
